package ca.ubc.cs.elide.crosscutting;

import ca.ubc.cs.elide.*;
import ca.ubc.cs.elide.utils.*;
import ca.ubc.cs.elide.nodes.*;

class PatternMatcher
{
	static Multimethod match = new Multimethod(PatternMatcher.class,"match",2);
	
	public static boolean matches(ChildBearingNode first, ParseNode second)
	{
		return (first.getClass().equals(second.getClass())) &&
			   ((Boolean)match.invoke(null, new Object[] {first,second})).
				   booleanValue();
	}
	
	public static boolean match(ClassNode first, ClassNode second)
	{
		String superPattern;
		if(first.getSuperclass().equals("Object"))
			superPattern = "@";
		else
			superPattern = first.getSuperclass();
		
		return  matchClassAncestor(first, second) &&
		        match(first.getPackage(), second.getPackage()) &&
			    match((ModifierBearingNode)first, (ModifierBearingNode)second) && 
			    InternalUtils.matchString(superPattern, second.getSuperclass()); 
	}
	
	public static boolean match(MethodNode first, MethodNode second)
	{
		return matchClassAncestor(first, second) &&
			   match((ModifierBearingNode)first, (ModifierBearingNode)second) && 
			   InternalUtils.matchString(first.getReturnType(), second.getReturnType());
	}
	
	public static boolean match(FieldNode first, FieldNode second)
	{
		return matchClassAncestor(first, second) &&
			   match((ModifierBearingNode)first, (ModifierBearingNode)second) && 
			   InternalUtils.matchString(first.getType(), second.getType());
	}
	
	public static boolean match(BlockNode first, BlockNode second)
	{
		return matchClassAncestor(first, second) &&
			   matchMethodAncestor(first, second) &&
			   match((ModifierBearingNode)first, (ModifierBearingNode)second);
	}
	
	public static boolean match(PackageNode first, PackageNode second)
	{
		return match((ParseNode)first, (ParseNode)second);
	}
	
	public static boolean match(ModifierBearingNode first, ModifierBearingNode second)
	{
		return match((ParseNode)first, (ParseNode)second) && 
			   (first.getModifiers() & second.getModifiers()) == first.getModifiers();
	}
	
	public static boolean match(ParseNode first, ParseNode second)
	{
		return InternalUtils.matchString(first.getName(), second.getName());
	}
	
	private static boolean matchClassAncestor(ChildBearingNode first, ChildBearingNode second)
	{
		return (first.getDeclaringClass() == null && second.getDeclaringClass() == null) || 
			   ((first.getDeclaringClass() != null && second.getDeclaringClass() != null) &&
				 match(first.getDeclaringClass(), second.getDeclaringClass()));
	}

	private static boolean matchMethodAncestor(BlockNode first, BlockNode second)
	{
		return (first.getMethod() == null && second.getMethod() == null) || 
			   ((first.getMethod() != null && second.getMethod() != null) &&
			     match(first.getMethod(), second.getMethod()));
	}
}