package ca.ubc.cs.elide.nodes;

import ca.ubc.cs.elide.*;
import ca.ubc.cs.elide.visitors.*;
import ca.ubc.cs.elide.utils.*;
import java.util.*;

/** A <code>ParameterNode</code> provides information about a single
 * formal parameter of a method.
 * <p>
 * When creating a <code>MethodNode</code> with <code>ClassNode.extend</code> or
  * <code>NodeFactory.createMethod</code>, these are automatically created and
  * added.
 **/
public class ParameterNode extends ModifierBearingNode
{
	private String type;
		
	public ParameterNode() {}

	public ParameterNode(String name, String type)
	{
		super(name);
		this.type = type;
	}
	
	/**
	 * Returns the name of the type of this parameter, as a String.
	 * <code>ClassNode.forName</code> can be used to retrieve a <code>ClassNode</code>
	 * for the type, if one is available.
	 **/
	public String getType() { return type; }
	
	/**
	 * Sets the name of the type of this parameter.
	 **/
	public void setType(String type) { this.type = type; }
	
	boolean isAddableTo(ChildBearingNode node)
	{
		return (node instanceof MethodNode);
	}
}