package ca.ubc.cs.elide.nodes;

import ca.ubc.cs.elide.*;
import ca.ubc.cs.elide.transformations.*;
import ca.ubc.cs.elide.visitors.*;
import ca.ubc.cs.elide.utils.*;
import ca.ubc.cs.elide.utils.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;

public abstract class ParseNode implements Cloneable, Serializable
{
	protected String name = "";
	protected ChildBearingNode parent;
	protected String comment = "";
		
	private Map typeMap = new HashMap();
	private Map propertyMap = new HashMap();
		
	public ParseNode() {}
	public ParseNode(String name) { this.name=name; }
	
	public String getName() { return name; }
	
	public void setName(String name)
	{
		this.name = name;
		if(parent != null)
			parent.indexChild(this);
	}

	/** Returns the comment associated with this item, as a string. **/
	public String getComment() { return comment; }
	
	/** Sets the comment associated with this item **/
	public void setComment(String comment) { this.comment = comment;}
	
	void setParent(ChildBearingNode parent)
	{
		if(isAddableTo(parent))
			this.parent = parent;
		else
		{
			System.out.println("Cannot add " + getClass() + " to " + parent.getClass());
			System.exit(1);
		}
	}
	
	public ChildBearingNode getParent() { return parent; }

	//these should probably be refactored for public consumption, but how?
	//what I sort of want is ClassNode.forVariable(varName, context);
	//but it needs to return a String, not a ClassNode
	public void registerType(String varName, String type)
	{
		typeMap.put(varName, type);
	}
	
	public String findType(String varName)
	{
		String type = (String) typeMap.get(varName);
		if(type == null && parent != null)
			type = parent.findType(varName);
		
		return type;
	}
			
	protected ParseNode getAncestor(Class ancestorClass) {
		if(ancestorClass.isInstance(parent) || parent == null)
			return parent;
		else
			return parent.getAncestor(ancestorClass);
	}
	
	/**
	 * Returns the <code>StartNode</code> object at the root of this parse tree.
	 **/
	public StartNode getRoot() { return (StartNode) getAncestor(StartNode.class); }
	
	/**
	 * Returns the <code>ClassNode</code> object representing the class within which this member was declared.
	 **/
	public ClassNode getDeclaringClass() { return (ClassNode) getAncestor(ClassNode.class); }
	
	/**
	 * Returns the <code>PackageNode</code> object representing the package within which this item was declared.
	 **/
	public PackageNode getPackage() { return (PackageNode) getAncestor(PackageNode.class); }
	
	public Object get(Object key) { return propertyMap.get(key); }
    public void put(Object key, Object prop) { propertyMap.put(key,prop); }
	
	/**
	 * Returns a representation of this object as Java source code.
	 **/
	public String toString()
	{
		StringWriter writer = new StringWriter();
		new Printer(writer).visit(this);
		return writer.toString();
	}
	
	public Object clone()
	{
		return InternalUtils.cloneObject(this);
	}
	
	boolean isAddableTo(ChildBearingNode node)
	{
		return false;
	}
}