package ca.ubc.cs.elide.sable;

import ca.ubc.cs.elide.*;
import ca.ubc.cs.elide.nodes.*;
import ca.ubc.cs.elide.transformations.*;

import j11.*;
import java.io.*;
import java.util.*;
import ca.mcgill.sable.gagnon.java.node.*;
import ca.mcgill.sable.gagnon.java.lexer.*;
import ca.mcgill.sable.gagnon.java.parser.*;
import ca.mcgill.sable.gagnon.java.analysis.*;

public class SableAdaptor extends PrettyPrinter
{
	protected StartNode root = new StartNode();
	protected ChildBearingNode index = root;
	protected List currentTransformArgs;
	protected CommentExtractor comments;
		
	public StartNode getRoot() { return root; }
	public void setRoot(StartNode root) { this.root = root; }
	
	/*** generic Java constructs ***/
	
	private List imports;
		
	public void caseACompilationUnit(ACompilationUnit node)
	{
		if(node.getPackageDeclaration() == null)
			index = ((StartNode)index).getDefaultPackage();
		else
			traverse(node.getPackageDeclaration());
		
		imports = node.getImportDeclaration();
		
		lineList(node.getTypeDeclaration());
		
		pop();
	}
	
	public void inAPackageDeclaration(APackageDeclaration node)
	{
		String name = prettyPrint(node.getName());
		PackageNode p = root.getPackage(name);
		if( p == null )
			push(new PackageNode(name));
		else index = p;
	}
	
	public void inASingleTypeImportDeclaration(ASingleTypeImportDeclaration node)
	{
		index.add(new ImportNode(prettyPrint(node.getName())));
	}
	
	public void inATypeImportOnDemandDeclaration(ATypeImportOnDemandDeclaration node)
	{
		String impt = prettyPrint(node.getName()) + ".*";
		
		index.add(new ImportNode(impt));
	}
	
	public void inAClassDeclaration(AClassDeclaration node)
	{
		String className = prettyPrint(node.getIdentifier());
		
		push(new ClassNode(className));
		getComment(node);
		
		lineList(imports);
	}
	
	public void outAClassDeclaration(AClassDeclaration node)
	{
		pop();
		resetBuffer();
	}
	
	public void inAInterfaceDeclaration(AInterfaceDeclaration node)
	{
		String ifaceName = prettyPrint(node.getIdentifier());
		ClassNode currentIface = new ClassNode(ifaceName);
		currentIface.setInterface(true);
		
		push(currentIface);
		getComment(node);
		
		lineList(imports);
	}
	
	public void outAInterfaceDeclaration(AInterfaceDeclaration node)
	{
		pop();
	}
	
	public void inASuper(ASuper node)
	{
		((ClassNode)index).setSuperclass(prettyPrint(node.getClassType()));
	}
	
	public void inAInterfaceTypeInterfaceTypeList(AInterfaceTypeInterfaceTypeList node)
	{
		index.add(0, new TypeNode(prettyPrint(node)));
	}
	
	public void inAInterfaceTypeListInterfaceTypeList(AInterfaceTypeListInterfaceTypeList node)
	{
		index.add(0, new TypeNode(prettyPrint(node.getInterfaceType())));
	}

	
	public void inAExtendsExtendsInterfaces(AExtendsExtendsInterfaces node)
	{
		index.add(0, new TypeNode(prettyPrint(node.getInterfaceType())));
	}
	
	public void inAExtendsInterfacesExtendsInterfaces(AExtendsInterfacesExtendsInterfaces node)
	{
		index.add(0, new TypeNode(prettyPrint(node.getInterfaceType())));
	}
	
	/*** fields - this gets tricky ***/
	
	private String fieldType;
	private String fieldName;
	private List fieldModifiers;
	
	public void caseAFieldDeclaration(AFieldDeclaration node)
	{
		fieldType = prettyPrint(node.getType());
		fieldModifiers = node.getModifier();
		
		traverse(node.getVariableDeclarators());
	}
	
	public void inALocalVariableDeclaration(ALocalVariableDeclaration node)
	{
		fieldType = prettyPrint(node.getType());
		fieldModifiers = null; //so we know we're in a var, not a field
	}
	
	public void inAVariableDeclaratorIdVariableDeclarator(AVariableDeclaratorIdVariableDeclarator node)
	{
		fieldName = prettyPrint(node);
		
		String localType = fieldType;
		if(fieldName.endsWith("[]"))
		{
			localType += "[]";
			fieldName = fieldName.substring(0, fieldName.length()-2);
		}
		
		index.registerType(fieldName, localType);
		if(fieldModifiers != null)
		{
			push(new FieldNode(fieldName, localType));
			wordList(fieldModifiers);
			pop();
		}
	}

	public void inAAssignVariableDeclarator(AAssignVariableDeclarator node)
	{
		fieldName = prettyPrint(node.getVariableDeclaratorId());

		String localType = fieldType;
		if(fieldName.endsWith("[]"))
		{
			localType += "[]";
			fieldName = fieldName.substring(0, fieldName.length()-2);
		}
		index.registerType(fieldName, localType);

		if(fieldModifiers != null)
		{
			push(new FieldNode(fieldName, localType));
			wordList(fieldModifiers);
			((FieldNode)index).setInitializer(prettyPrint(node.getVariableInitializer()));
			pop();
		}
	}
	
	/*** methods ***/
	
	public void inAMethodDeclaration(AMethodDeclaration node)
	{
		push(new MethodNode(""));
		getComment(node);
	}
	
	public void outAMethodDeclaration(AMethodDeclaration node)
	{
		pop();
	}

	public void inAAbstractMethodDeclaration(AAbstractMethodDeclaration node)
	{
		MethodNode currentMethod = new MethodNode("");
		push(currentMethod);
		getComment(node);
	}
	
	public void outAAbstractMethodDeclaration(AAbstractMethodDeclaration node)
	{
		pop();
	}
	
	public void inAIdentifierMethodDeclarator(AIdentifierMethodDeclarator node)
	{
		index.setName(prettyPrint(node.getIdentifier()));
	}
	
	public void inATypeMethodHeader(ATypeMethodHeader node)
	{
		String returnType = prettyPrint(node.getType());
		
		((MethodNode)index).setReturnType(returnType);
	}
	
	public void inAVoidMethodHeader(AVoidMethodHeader node)
	{
		((MethodNode)index).setReturnType("void");
	}
	
	public void inAClassTypeListClassTypeList(AClassTypeListClassTypeList node)
	{
		String type = prettyPrint(node.getClassType());
		index.add(0, new TypeNode(type));
	}
	
	public void inAClassTypeClassTypeList(AClassTypeClassTypeList node)
	{
		String type = prettyPrint(node.getClassType());
		index.add(0, new TypeNode(type));
	}
	
	public void inAFormalParameter(AFormalParameter node)
	{
		String name = prettyPrint(node.getVariableDeclaratorId());
		String type = prettyPrint(node.getType());
		
		if(name.endsWith("[]"))
		{
			type += "[]";
			name = name.substring(0, name.length()-2);
		}		
		push(new ParameterNode(name, type));
	}
	
	public void outAFormalParameter(AFormalParameter node)
	{
		pop();
	}
	
	public void inASemicolonMethodBody(ASemicolonMethodBody node)
	{
	}
	
	public void inABlockMethodBody(ABlockMethodBody node)
	{
		resetBuffer();
		resetIndent();
		resetBlock();
	}
	
	public void outABlockMethodBody(ABlockMethodBody node)
	{
		index.add(new TextNode(getBuffer()));
	}
	
	public void inAConstructorDeclaration(AConstructorDeclaration node)
	{
		push(new ConstructorNode());
		getComment(node);
	}
	
	public void outAConstructorDeclaration(AConstructorDeclaration node)
	{
		pop();
	}
	
	public void inAConstructorDeclarator(AConstructorDeclarator node)
	{
		index.setName(prettyPrint(node.getSimpleName()));
	}
		
	public void inAConstructorBody(AConstructorBody node)
	{
		resetBuffer();
		resetIndent();
	}
	
	public void outAConstructorBody(AConstructorBody node)
	{
		index.add(new TextNode(getBuffer()));
	}
	
	public void caseAStaticInitializer(AStaticInitializer node)
	{
		push(new StaticBlockNode());
		
		resetBuffer();
		resetIndent();
		resetBlock();
	
		traverse(node.getBlock());
		index.add(new TextNode(getBuffer()));

		pop();
	}
	
	public void caseABlockClassBodyDeclaration(ABlockClassBodyDeclaration node)
	{
		push(new InstanceBlockNode());
		
		resetBuffer();
		resetIndent();
		resetBlock();
		
		traverse(node.getBlock());
		index.add(new TextNode(getBuffer()));
		
		pop();
	}
	
	/*** code blocks ***/
	
	private boolean firstBlock = false;		
	protected void resetBlock()
	{
		firstBlock = true;
	}
	
	public void inABlock(ABlock node)
	{
		if(firstBlock)
			firstBlock = false;
		else
			super.inABlock(node);
	}
	
	public void outABlock(ABlock node)
	{
		if(getIndent() > 0)
			super.outABlock(node);
	}
	
	/*** intercept things we can't deal with yet ***/
	
	public void caseASimpleClassInstanceCreationExpression(ASimpleClassInstanceCreationExpression node) {
		print(prettyPrint(node));
	}
	public void caseAQualifiedClassInstanceCreationExpression(AQualifiedClassInstanceCreationExpression node) {
		print(prettyPrint(node));
	}
	public void caseACatchClause(ACatchClause node) {
		print(prettyPrint(node));
	}
	
	protected String prettyPrint(Object node)
	{
		PrettyPrinter printer = new PrettyPrinter();
		((Node) node).apply(printer);
		return printer.getBuffer().trim();
	}
	
	/*** DDSL-specific syntax ***/
	
	public void inADdslModifier(ADdslModifier node)
	{
		currentTransformArgs = new ArrayList();
	}
	
	public void outADdslModifier(ADdslModifier node)
	{
		index.add(new TransformNode(prettyPrint(node.getName()), currentTransformArgs));
	}
	
	public void inANameDdslArgument(ANameDdslArgument node)
	{
		currentTransformArgs.add(prettyPrint(node));
	}
	
	public void inAStringLiteralDdslArgument(AStringLiteralDdslArgument node)
	{
		currentTransformArgs.add(prettyPrint(node));
	}
	
	public void inADdslInitializerDdslArgument(ADdslInitializerDdslArgument node)
	{
		String text = prettyPrint(node);
		text = text.substring(2, text.length() - 2);
		currentTransformArgs.add(text);
	}
	
	public void inAExpressionDdslExprArgument(AExpressionDdslExprArgument node)
	{
		currentTransformArgs.add(prettyPrint(node));
	}
	
	public void inAFormalParameterDdslExprArgument(AFormalParameterDdslExprArgument node)
	{
		currentTransformArgs.add(prettyPrint(node));
	}
	
	public void caseADdslExpr(ADdslExpr node)
	{
		index.add(new TextNode(getBuffer()));
		
		currentTransformArgs = new ArrayList();
		traverse(node.getDdslExprArguments());
		
		push(new BlockNode());
		index.add(new TransformNode(
			prettyPrint(node.getDdslExprIdentifier()).substring(1),
			currentTransformArgs));
		
		int indent = resetIndent();
		resetBuffer();
		resetBlock();
		traverse(node.getBlock());
		index.add(new TextNode(resetBuffer()));
		setIndent(indent);
		
		pop();
	}
	
	public void caseADdslBlock(ADdslBlock node)
	{
		if(index instanceof ClassNode)
			push(new InstanceBlockNode());
		else
		{
			index.add(new TextNode(getBuffer()));
			push(new BlockNode());
		}
		
		traverse(node.getDdslModifier());
		
		int indent = resetIndent();
		resetBuffer();
		resetBlock();
		
		traverse(node.getBlock());
		
		index.add(new TextNode(resetBuffer()));
		setIndent(indent);
		
		pop();
	}
	
	/*** modifiers ***/

	public void inAPublicModifier(APublicModifier node) { ((ModifierBearingNode)index).makePublic(); }
	public void inAProtectedModifier(AProtectedModifier node) {((ModifierBearingNode)index).makeProtected(); } 
	public void inAPrivateModifier(APrivateModifier node) { ((ModifierBearingNode)index).makePrivate();}

	public void inAAbstractModifier(AAbstractModifier node) { ((ModifierBearingNode)index).setAbstract(true); }
	public void inAFinalModifier(AFinalModifier node) { ((ModifierBearingNode)index).setFinal(true); }
	public void inANativeModifier(ANativeModifier node) { ((ModifierBearingNode)index).setNative(true); }
	public void inATransientModifier(ATransientModifier node) { ((ModifierBearingNode)index).setTransient(true); }
	public void inAVolatileModifier(AVolatileModifier node) { ((ModifierBearingNode)index).setVolatile(true); }
	public void inAStaticModifier(AStaticModifier node) { ((ModifierBearingNode)index).setStatic(true); }
	public void inASynchronizedModifier(ASynchronizedModifier node) { ((ModifierBearingNode)index).setSynchronized(true); }
	
	/** utils ***/
	
	protected void push(ChildBearingNode node)
	{
		index.add(node);
		index = node;
	}
	
	protected void pop()
	{
		index = index.getParent();
	}
	
	protected void getComment(Node node)
	{
		index.setComment(comments.getComment(node));
	}
	
	/*** compilation ***/

	public void compile(String file)
	{
       try
        {
            UnicodePreprocessor preprocessor = new UnicodePreprocessor(
                new PushbackReader(
                new BufferedReader(
                new FileReader(file)), 1024));

            Lexer lexer = new Lexer(new PushbackReader(preprocessor, 1024));
            Parser parser = new Parser(lexer);
            Start ast = parser.parse();
			
			comments = new CommentExtractor(parser.ignoredTokens);
			
			ast.apply(this);
        }
		catch(ca.mcgill.sable.gagnon.java.parser.ParserException e)
		{
			System.out.println("Error parsing " + file);
			System.out.println("at token " + e.getToken() + " " + e.getMessage());
			System.exit(1);
		}
   		catch(ca.mcgill.sable.gagnon.java.lexer.LexerException e)
		{
			System.out.println("Error lexing " + file + ": " + e.getMessage());
            System.exit(1);
        }
       catch(IOException e)
        {
            e.printStackTrace();
            System.out.println(e);
            System.exit(1);
        }
   	}
	
	public void addCompiledText(String text)
	{
		try
		{
			UnicodePreprocessor preprocessor = new UnicodePreprocessor(
				new PushbackReader(
				new BufferedReader(
				new StringReader(text)), 1024));
			
			Lexer lexer = new Lexer(new PushbackReader(preprocessor, 1024));
			Parser parser = new Parser(lexer);
			Start ast = parser.parse();
			
			comments = new CommentExtractor(parser.ignoredTokens);
			
			ast.apply(this);
		}
		catch(ca.mcgill.sable.gagnon.java.parser.ParserException e)
		{
			System.out.println("Error: " + e.getMessage() + " while parsing ");
			System.out.println(text);
			System.exit(1);
		}
  		catch(ca.mcgill.sable.gagnon.java.lexer.LexerException e)
		{
			System.out.println("Error: " + e.getMessage() + " while lexing ");
			System.out.println(text);
            System.exit(1);
        }
		catch(IOException e)
		{
			System.out.println("BUG: IOException while parsing " + e);
		}
	}
	
	public static StartNode compileText(String text)
	{
		SableAdaptor adaptor = new SableAdaptor();
		adaptor.addCompiledText(text);
		return adaptor.getRoot();
	}
}