package ca.ubc.cs.elide.transformations;

import java.util.*;

public class Executor
{
	private Map status = new HashMap();
	private Map unsatisfied = new HashMap();
	private Set transformations;
	private List resolvedTransformations;
		
	public Executor(Set dependencies)
	{
		transformations = new HashSet(dependencies);
	}
	
	public List resolveAll()
	{
		resolvedTransformations = new ArrayList();
		registerAll();
		executeAll();
		return resolvedTransformations;
	}
		
	private void executeAll()
	{
		Iterator i = transformations.iterator();
		while(i.hasNext()) execute((Dependency)i.next());
	}
	
	private void registerAll()
	{
		Iterator i = transformations.iterator();
		while(i.hasNext()) register((Dependency)i.next());
	}
	
	private void register(Dependency transform)
	{
		Iterator satisfies = transform.getSatisfies().iterator();
		while(satisfies.hasNext())
		{
			String name = (String) satisfies.next();
			List unsatisfiedTransforms = (List) unsatisfied.get(name);
			if(unsatisfiedTransforms == null)
			{
				unsatisfiedTransforms = new ArrayList();
				unsatisfied.put(name, unsatisfiedTransforms);
			}
			
			unsatisfiedTransforms.add(transform);
		}
	}
	
	private void execute(Dependency transform)
	{
		if(isDone(transform))
			return;
		else if(isInProgress(transform))
			circularDependency(transform);
		
		
		inProgress(transform);
		
		Iterator dependencies = transform.getRequires().iterator();
		while(dependencies.hasNext())
		{
			String dependency = (String) dependencies.next();
			List dependants = (List) unsatisfied.get(dependency);
			
			if(dependants != null)
			{
				Iterator i = dependants.iterator();
				while(i.hasNext())
					execute((Dependency) i.next());

				unsatisfied.remove(dependency);
			}
		}
		
		resolvedTransformations.add(transform);
		
		done(transform);
	}
	
	/** status markers ***/
	
	static final String DONE = "done";
	static final String INPROGRESS = "in progress";
	private boolean isDone(Dependency transform)
	{
		return status.get(transform) == DONE;
	}
	private boolean isInProgress(Dependency transform)
	{
		return status.get(transform) == INPROGRESS;
	}
	private void inProgress(Dependency transform)
	{
		status.put(transform, INPROGRESS);
	}
	private void done(Dependency transform)
	{
		status.put(transform, DONE);
	}
	
	/*** error ***/
	
	private void circularDependency(Dependency transform)
	{
		System.out.println("Error: " + transform + " depends on itself");
		System.out.println("Active transformations: ");
		Iterator i = status.entrySet().iterator();
		while(i.hasNext())
		{
			Map.Entry entry = (Map.Entry) i.next();
			if(entry.getValue() == INPROGRESS)
				System.out.println(entry.getKey());
		}
		System.exit(1);
	}
}