package ca.ubc.cs.elide.transforms;

import ca.ubc.cs.elide.nodes.*;
import ca.ubc.cs.elide.*;
import ca.ubc.cs.elide.transformations.*;
import java.util.*;

public class DependencySet extends Transform
{
        public void transform(MethodNode target, String satisfiesOrRequires)
        {
        	ClassNode parentClass = target.getDeclaringClass();
		
			addDepSet(parentClass);
			addDepInstance(parentClass, target.getName());
                
			String action;
			if(getClass() == Satisfies.class)
				action = "Satisfies";
			else
				action = "Requires";
			
			parentClass.extend("static { " + target.getName() + "Dependency.add" + action + "(\""
				+ satisfiesOrRequires + "\"); }");                              
        }
		
        public void addDepSet(ClassNode klass)
        {
 			if(klass.getDeclaredField("dependencySet") == null)
            {
                klass.extend("private static java.util.Set dependencySet = new java.util.HashSet();");
				                        
				klass.extend("public static java.util.Set getDependencySet(Class self)"+
		        "	      	  {"+
		        "             	java.util.Set depSet = new java.util.HashSet();"+
		        "             	depSet.addAll(dependencySet);"+
				"             	java.util.Set superDepSet = "+superDependencies(klass)+";"+
				"             	java.util.Iterator i = superDepSet.iterator();"+
				"             	while( i.hasNext() )"+
				"             	{"+
				"             		ca.ubc.cs.elide.transformations.Dependency d = (ca.ubc.cs.elide.transformations.Dependency)((ca.ubc.cs.elide.transformations.Dependency)i.next()).clone();"+
				"             		d.setTransformName(\""+klass.getShortName()+"\");"+
				"	             	depSet.add(d);"+
				"               }"+
				"	        	return depSet;"+
				"             }");
			}
        }
		
		private String superDependencies(ClassNode klass)
		{
			String superDependencies;
			String superclass = klass.getSuperclass();
			
			if( !superclass.equals("Transform") )
			{
				superDependencies = "Transform.getDependencies(\""+superclass+"\")";
			}
			else superDependencies = "new java.util.HashSet()";
			return superDependencies;
		}
        
        public void addDepInstance(ClassNode klass, String name)
        {
				//should this be getField() ?
        	if(klass.getDeclaredField(name + "Dependency") == null)
            {
                klass.extend("private static ca.ubc.cs.elide.transformations.Dependency " +
            		name + "Dependency = new ca.ubc.cs.elide.transformations.Dependency(\"" + name + "\", \"" + klass.getShortName() +"\");");
                klass.extend("static { dependencySet.add(" + name + "Dependency); }");
            }
        }
}