package ca.ubc.cs.elide.utils;

import ca.ubc.cs.elide.nodes.*;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;
import org.apache.oro.text.perl.*; 

public class InternalUtils
{
	static String transformSuffix = ".ddsl";
	static String aspectSuffix = ".aspect";
	
	public static boolean endsWithTransformSuffix(String fileName)
	{
		return fileName.endsWith(transformSuffix);
	}
	
	public static boolean endsWithAspectSuffix(String fileName)
	{
		return fileName.endsWith(aspectSuffix);
	}
	
	public static String getPath(String packageName)
	{
		return packageName.replace('.',File.separatorChar);
	}
	
	static Perl5Util perl = new Perl5Util();

	public static String quote(String text)
	{
		text = text.substring(2, text.length() - 2);
		PushbackReader reader = new PushbackReader(new StringReader(text));
		Writer writer = new StringWriter();
		PrintWriter printer = new PrintWriter(writer);

		printer.print("\"");
		try
		{
			quote_text(reader, new PrintWriter(writer));
		}
		catch(IOException e)
		{}
		printer.print("\"");
		
		return writer.toString();
	}

	/**somewhat stolen from Brad Cox's MLS */
	
	private final static String lineSeparator = System.getProperty("line.separator");	
	private static void quote_text(PushbackReader in, PrintWriter out)
		throws IOException
	{
		int thisChar, nextChar;
		while((thisChar = in.read()) != -1)
		{
			switch(thisChar)
			{
				case '\r': 
					break;
				case '\n':
				  out.print("\\n\""+lineSeparator+"+\"");
				  break;
				case '\\':
				  out.print("\\" + (char)thisChar);
				  break;
				case '"':
				  out.print("\\\"");
				  break;
				case '#':
				  nextChar = in.read();
				  if (nextChar == '{')
				  {
					out.print("\"+");
					quote_code(in, out);
					out.print("+\"");
					break;
				  }
				  else 
				  {
					out.print((char)thisChar);
					in.unread(nextChar);
					break;
				  }
				case '}':
				  nextChar = (char)in.read();
				  if (nextChar == '%')
				  {
					out.print("\"");
					return;
				  }
				  else 
				  {
					out.print((char)thisChar);
					in.unread((char)nextChar);
					break;
				  }
				default:
				  out.print((char)thisChar);
				  break;
			}
		}
	}
	
	private static void quote_code(PushbackReader in, PrintWriter out)
		throws IOException
	{
		int thisChar, nextChar;
		while((thisChar = in.read()) != -1)
		{
			switch(thisChar)
			{
				case '\r':
					break;
				case '\n':
					out.print(lineSeparator);
					break;
				case '%':
					nextChar = in.read();
					if (nextChar == '{')
					{
						quote_text(in, out);
						break;
					}
					else 
					{
						out.print((char)thisChar);
						in.unread(nextChar);
						break;
					}
					case '}':
						nextChar = (char)in.read();
						if (nextChar == '#')
							return;
						else 
						{
							out.print((char)thisChar);
							in.unread((char)nextChar);
							break;
						}
					default:
						out.print((char)thisChar);
						break;
			  }
		  }
	}
	
	public static Object cloneObject(Object obj)
	{
		try
		{
			ByteArrayOutputStream baos = new ByteArrayOutputStream(); 
			ObjectOutputStream out = new ObjectOutputStream(baos); 
			out.writeObject(obj); 
			ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray()); 
			ObjectInputStream in = new ObjectInputStream(bais); 
			return(in.readObject()); 
		}
		catch(IOException e)
		{
			throw new RuntimeException("IO error in cloneObject().");
		}
		catch(ClassNotFoundException e)
		{
			throw new RuntimeException("Class not found in cloneObject().");
		}
	}
	
	public static boolean matchString(String pattern, String other)
	{
		return perl.match("/^" + perl.substitute("s/@/.+/g", pattern) + "$/", other);
	}
	
	public static boolean matchString(String pattern, List others)
	{
		Iterator i = others.iterator();
		while( i.hasNext() )
			if( matchString(pattern, (String)i.next()) )
				return true;
		return false;
	}
}