package ca.ubc.cs.elide.utils;

import ca.ubc.cs.elide.visitors.*;
import ca.ubc.cs.elide.sable.*;
import ca.ubc.cs.elide.nodes.*;
import java.io.*;
import org.apache.oro.text.perl.*; 

public class SourceFile
{
	private static String parseDir = "parses";
	String fileName;
	Perl5Util perl = new Perl5Util();
	
	public String parseFileName() 
	{	
		return parseDir+File.separator+mangle(fileName)+".parse";
	}
	
	private String mangle(String text)
	{
		return perl.substitute("s/:/___/", text);
	}
	
	public SourceFile(String fileName)
	{
		String mangled = mangle(fileName);
		this.fileName = fileName;
		String path = mangled.substring(0,mangled.lastIndexOf(File.separator)+1);
		(new File(parseDir+File.separator+path)).mkdirs();			
	}
	
	public void load(Merger merger)
	{
		long lastModified = (new File(fileName)).lastModified();
		File parseFile = new File(parseFileName());
		
		if(parseFile.exists() && lastModified <= parseFile.lastModified())
		{
			System.out.println("Loading " + parseFileName() + "...");
			merger.visit(read(parseFileName()));
		}
		else
		{
			SableAdaptor parser = new SableAdaptor();
			System.out.println("Parsing " + fileName + "...");
			parser.compile(fileName);
			StartNode newTree = parser.getRoot();
			write(newTree, parseFileName());
			merger.visit(newTree);
		}
	}
	
	public StartNode read(String inFileName) 
	{
		try
		{
			FileInputStream fis = new FileInputStream(new File(inFileName)); 
			ObjectInputStream in = new ObjectInputStream(fis); 
			return (StartNode)in.readObject(); 
		}
		catch(IOException e)
		{
			System.out.println(e);
			throw new RuntimeException("IO error in read(): "+inFileName);
		}
		catch(ClassNotFoundException e)
		{
			System.out.println(e);
			throw new RuntimeException("Class not found in read(): "+inFileName);
		}
	}
	
	public void write(StartNode tree, String outFileName)
	{
		try
		{
			File outFile = new File(outFileName);
			outFile.createNewFile();
			FileOutputStream fos = new FileOutputStream(outFile); 
			ObjectOutputStream out = new ObjectOutputStream(fos); 
			out.writeObject(tree); 
		}
		catch(IOException e)
		{
			System.out.println(e);
			throw new RuntimeException("IO error in write(): "+outFileName);
		}
	}
}